/*
 * Copyright (c) 2018, 2019, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#ifndef SHARE_MEMORY_METASPACE_CHUNKTREE_HPP
#define SHARE_MEMORY_METASPACE_CHUNKTREE_HPP

#include "memory/allocation.hpp"
#include "memory/metaspace/abstractPool.hpp"
#include "memory/metaspace/chunkLevel.hpp"
#include "memory/metaspace/metachunk.hpp"

namespace metaspace {

// Chunks live in a binary tree.
//

class ChunkClosure {
 public:
  // Return false to cancel traversal.
  virtual bool do_chunk(Metachunk* chunk) = 0;
};


class ChunkTree : public CHeapObj<mtClass> {

  typedef u2 ref_t;

  // Root is either a reference to a chunk (in that case, a root chunk of max. size)
  // or to a node.
  ref_t _root;

  struct btnode_t {

    ref_t parent;
    ref_t child[2];

  };

  typedef AbstractPool<btnode_t, ref_t> NodePoolType;
  typedef AbstractPool<Metachunk, ref_t> ChunkPoolType;
  NodePoolType _nodePool;
  ChunkPoolType _chunkPool;

  // The highest bits of a reference encodes the type:
  // bit 15:  0 - reference is a btnode_t
  //          1 - reference is a chunk
  // This also means a reference has to get by with 15 bits. Which covers 32K, which is enough for both
  // chunk headers and nodes within one root chunk area.
  static const u2 highest_possible_index_value = 0x7FFF;

  static u2 get_raw_index_from_reference(ref_t ref)     { return 0x7FFF & ref; }

  static bool reference_is_node(ref_t ref)              { return 0x8000 & ref == 0; }
  static ref_t make_node_reference(u2 raw_idx) {
    assert(raw_idx <= highest_possible_index_value, "invalid index");
    return raw_idx;
  }

  static bool reference_is_chunk(ref_t ref)             { return 0x8000 & ref != 0; }
  static ref_t make_chunk_reference(u2 raw_idx) {
    assert(raw_idx <= highest_possible_index_value, "invalid index");
    return 0x8000 | raw_idx;
  }

#ifdef ASSERT
  void check_is_valid_node_ref(ref_t ref) const;
  void check_is_valid_chunk_ref(ref_t ref) const;
#endif

  // Given a reference we assume to be a node, resolve it to the node pointer.
  // Returns 0 if the reference is not a node ref but a chunk ref.
  btnode_t* resolve_reference_to_node_or_null(ref_t ref) const {
    if (reference_is_node(ref)) {
      return _nodePool.elem_at_index(get_raw_index_from_reference(ref));
    }
    return NULL;
  }

  // Given a reference we know to be a node, resolve it to the node pointer.
  btnode_t* resolve_reference_to_node(ref_t ref) const {
    btnode_t* n = resolve_reference_to_node_or_null(ref);
    assert(n != NULL, "Not a node ref.");
    return n;
  }

  // Allocate a new node. Node is uninitialized.
  // Returns pointer to node, and reference in ref.
  btnode_t* allocate_new_node() {
    return _nodePool.allocate_element();
  }

  // Given a node pointer, return its correctly encoded reference.
  ref_t encode_reference_for_node(const btnode_t* n) const {
    const u2 raw_idx = _nodePool.index_for_elem(n);
    return make_node_reference(raw_idx);
  }

  // Release a node to the pool.
  void release_node(btnode_t* n) {
    _nodePool.return_element(n);
  }

  // Given a reference we assume is a chunk reference, resolve it to the chunk pointer.
  // Returns 0 if reference is not a chunk reference (but a node reference).
  Metachunk* resolve_reference_to_chunk_or_null(ref_t ref) const {
    if (reference_is_chunk(ref)) {
      return _chunkPool.elem_at_index(get_raw_index_from_reference(ref));
    }
    return NULL;
  }

  // Given a reference we know to be a chunk, resolve it to the chunk pointer.
  Metachunk* resolve_reference_to_chunk(ref_t ref) const {
    Metachunk* c = resolve_reference_to_chunk_or_null(ref);
    assert(c != NULL, "Not a chunk ref.");
    return c;
  }

  // Allocate a new node. Node is uninitialized.
  // Returns pointer to node, and reference in ref.
  Metachunk* allocate_new_chunk() {
    return _chunkPool.allocate_element();
  }

  // Given a chunk pointer, return its correctly encoded reference.
  ref_t encode_reference_for_chunk(Metachunk* c) const {
    const u2 raw_idx = _chunkPool.index_for_elem(c);
    return make_chunk_reference(raw_idx);
  }

  // Release a chunk to the pool.
  void release_chunk(Metachunk* c) {
    _chunkPool.return_element(c);
  }

  //// Helpers for tree traversal ////

  class ConstChunkClosure;
  bool iterate_chunks_helper(ref_t ref, ChunkClosure* cc) const;

#ifdef ASSERT
  // Verify a life node (one which lives in the tree).
  void verify_node(const btnode_t* n) const;
  // Helper for verify()
  void verify_helper(bool slow, ref_t ref, const MetaWord* p, int* num_chunks, int* num_nodes) const;
#endif

  // Given a chunk c, split it once.
  //
  // The original chunk must not be part of a freelist.
  //
  // Returns pointer to the result chunk; updates the splinters array to return the splintered off chunk.
  //
  // Returns NULL if chunk cannot be split any further.
  Metachunk* split_once(Metachunk* c, Metachunk* splinters[chklvl::NUM_CHUNK_LEVELS]);

  // Given a chunk, attempt to merge it with its sibling if it is free.
  // Returns pointer to the result chunk if successful, NULL otherwise.
  //
  // Returns number of merged chunks, by chunk level, in num_merged array. These numbers
  // includes the original chunk.
  //
  // !!! Please note that if this method returns a non-NULL value, the
  // original chunk will be invalid and should not be accessed anymore! !!!
  Metachunk* merge_once(Metachunk* c, int num_merged[chklvl::NUM_CHUNK_LEVELS]);

public:

  ChunkTree();
  virtual ~ChunkTree() {}

  // Initialize: allocate a root node and a root chunk header; return the
  // root chunk header. It will be partly initialized.
  // Note: this just allocates a memory-less header; memory itself is allocated inside VirtualSpaceNode.
  Metachunk* alloc_root_chunk_header();

  // Given a chunk c, split it recursively until you get a chunk of the given target_level.
  //
  // The original chunk must not be part of a freelist.
  //
  // Returns pointer to the result chunk; returns split off chunks in splinters array.
  //
  // Returns NULL if chunk cannot be split at least once.
  Metachunk* split(chklvl_t target_level, Metachunk* c, Metachunk* splinters[chklvl::NUM_CHUNK_LEVELS]);

  // Given a chunk, attempt to merge it recursively with its neighboring chunks.
  //
  // If successful (merged at least once), returns address of
  // the merged chunk; NULL otherwise.
  //
  // The merged chunks are removed from their freelist; the number of merged chunks is
  // returned, split by level, in num_merged array. Note that these numbers does not
  // include the original chunk.
  //
  // !!! Please note that if this method returns a non-NULL value, the
  // original chunk will be invalid and should not be accessed anymore! !!!
  Metachunk* merge(Metachunk* c, int num_merged[chklvl::NUM_CHUNK_LEVELS]);

  //// tree traversal ////

  // Iterate over all nodes in this tree. Returns true for complete traversal,
  // false if traversal was cancelled.
  bool iterate_chunks(ChunkClosure* cc) const;


  //// Debug stuff ////

  // Verify tree. If base != NULL, it should point to the location assumed
  // to be base of the first chunk.
  DEBUG_ONLY(void verify(bool slow, const MetaWord* base) const;)

  // Returns the footprint of this tree, in words.
  size_t memory_footprint_words() const;


};


///////////////////////
// An C-heap allocated array of chunk trees. Used to describe fragmentation over a range of multiple root chunks.
class ChunkTreeArray {

  const MetaWord* const _base;
  const size_t _word_size;

  ChunkTree** _arr;
  int _num;

#ifdef ASSERT
  void check_pointer(const MetaWord* p) const {
    assert(p >= _base && p < _base + _word_size, "Invalid pointer");
  }
#endif

  int index_by_address(const MetaWord* p) const {
    DEBUG_ONLY(check_pointer(p);)
    return (p - _base) / chklvl::MAX_CHUNK_WORD_SIZE;
  }

public:

  // Create an array of ChunkTree objects, all initialized to NULL, covering
  // a given memory range. Memory range must be aligned to size of root chunks.
  ChunkTreeArray(const MetaWord* base, size_t word_size);

  ~ChunkTreeArray();

  // Given a memory address into the range the trees cover, return the corresponding
  // tree. If none existed at this position, create it.
  ChunkTree* get_tree_by_address(const MetaWord* p) const {
    assert(p >= _base && p < _base + _word_size, "Invalid pointer");
    const int idx = index_by_address(p);
    assert(idx >= 0 && idx < _num, "Invalid index");
    if (_arr[idx] == NULL) {
      _arr[idx] = new ChunkTree();
    }
    return _arr[idx];
  }

  // Iterate over all nodes in all trees. Returns true for complete traversal,
  // false if traversal was cancelled.
  bool iterate_chunks(ChunkClosure* cc) const;

  DEBUG_ONLY(void verify(bool slow) const;)

  // Returns the footprint of all trees in this array, in words.
  size_t memory_footprint_words() const;

};


} // namespace metaspace

#endif // SHARE_MEMORY_METASPACE_CHUNKTREE_HPP
